"""
This module is modified from https://github.com/locuslab/wanda.

It contains utility functions and classes for pruning neural network models using the Wanda method.
The WANDA method is a weight pruning technique that aims to reduce the number of parameters in a neural network
while maintaining its performance.
"""
